#! /usr/bin/perl -w

# antihtml

# Generate a bounce message for an email on stdin, and post it using exim.
# Usage: antihtml <return-path> <failed-recipient>...

# (c) Darren Salt

#$version = '1.2.0, 25 June 2001';

sub scanbody ($;$@);
sub getHeader ($;$);

use POSIX qw(strftime);

###
# Configurable stuff goes here

# Mail software
$mailer = 'exim -bm -t -f"<>"';

# What to use to log bounces
@logfile = ('/usr/bin/logger', '-p', 'mail.warn');

# List of content types to "bounce"
# List of MIME types (entered as partial regexps)
#   Primary content types (*not* the multipart types!)
#   KEEP IN SYNC WITH THE 'contains' CONDITIONS IN antihtml !
@primary = (
  'text/html',
  'text/richtext',
  'application/msword'
);
#   multipart/mixed (should normally be empty)
@mixed = ();
#   multipart/alternative (e.g. HTML, Word)
@alternative = (
  'text/html',
  'application/msword',
);
#   multipart/related (consolidated attachments, formatting info)
@related = (
  'application/ms-tnef'
);

# List of descriptions (referenced by the above description keys)
%describe = (
  'text/html',			'HTML content',
  'text/richtext',		'RTF content',
  'application/ms-tnef',	'consolidated formatting/attachment data',
  'application/msword',		'proprietary-format content',
  '',				'unrecognised but unacceptable content'
);

# End of configurable stuff
###

# Environment and arguments
$ENV{'PATH'} = '/usr/sbin';
$from = shift @ARGV;
@received = @ARGV;
@ARGV = ();
$| = 1;

# Must have at least two arguments...
die unless defined $from && defined @received;

# Sort out the recipients list
@recipients = ();
foreach $i (@received) {
  push @recipients, split /, /, $i;
}

# What's my mailname?
open HOST, '< /etc/mailname' or open HOST, '< /etc/hostname' or die $!;
$mailname = <HOST>;
close HOST;
die unless defined $mailname;
chomp $mailname;

$to = $from;
$re_primary = '(?-xism:^('.join('|', @primary).')\b)' if $#primary >= 0;
$re_secondary{'alternative'} = '(?-xism:^('.join('|', @alternative).')\b)' if $#alternative >= 0;
$re_secondary{'related'} = '(?-xism:^('.join('|', @related).')\b)' if $#related >= 0;
$re_secondary{'mixed'} = '(?-xism:^('.join('|', @mixed).')\b)' if $#mixed >= 0;
$re_multipart = '(?-xism:^multipart\/(mixed|alternative|related)\b)';

# Slurp in the headers
$headers = '';
while ($line = <>) {
  last if $line =~ /^\r?\n?$/o;
  $headers .= '  '.$line;
}

# Do we really have a Content-Type header?
$mime = getHeader ('Content-Type');
exit 0 unless defined $mime;

# Read the aliases file into a hash
%aliases = ();
open ALIAS, '< /etc/aliases';
while ($line = <ALIAS>) {
  chomp $line;
  next if $line =~ /^#/o;
  $line =~ /^(.+):[ \t]*([^ \t]+)/;
  $aliases{$1} = $2 if defined $2;
}
close ALIAS;

# Make a list of local IDs which exist and have home directories
# from the recipients list - we don't want to generate *two* bounces
# for some recipients...!
%bounce = ();
foreach $i (@recipients) {
  $i =~ /^(.*)@([^@]*)$/o;
  local $account = $1;
  local $forward = $2;
  $forward =~ tr/A-Z/a-z/;
  next unless $forward eq $mailname;
  $account = $aliases{$account} if defined $aliases{$account};
  local @pwent = getpwnam ($account);
  $bounce{$i} = '' if defined @pwent && -d $pwent[7];
}

# Build part of the bounce msg from the above list of local IDs
$addresses = '';
foreach $i (keys %bounce) {
  $addresses .= "\n  ".$i;
}
# Stop now if it's empty
exit 0 if $addresses eq '';

# Parse the headers of each chunk of the message (if multipart)
# Exit if nothing interesting is found
undef $line;
if ($mime =~ $re_multipart) {
  $line = scanbody ($mime);
} elsif (defined $re_primary) {
  $mime =~ $re_primary;
  $line = $1;
}

exit 0 unless defined $line && $line ne '';
# $line now holds the interesting MIME type

# We have a valid recipient *and* the message contains HTML
$addresses = substr ($addresses, 1);
$addrlist = substr ($addresses, 2);
$addrlist =~ s/  /\t/g;
$line = '' unless defined $describe{$line};

# Now let's write the bounce...
open EXIM, '| '.$mailer or die $!;
print EXIM <<EOF or die $!;
X-Failed-Recipients: $addrlist
To: $to
Subject: Mail delivery failed: returning message to postmaster

This message was created automatically by mail delivery software.

A message that you sent could not be delivered to one or more of its
recipients due to $describe{$line}.

You are strongly advised to select the 'plain text' option in your
mail software.

The following address(es) failed:

$addresses

----- Headers follow -----

$headers
EOF
close EXIM or die $!;

# Log the fact that we've "rejected" a message
$line = getHeader('Message-ID');
$line = '(null ID)' unless defined $line && $line ne '';
push @logfile, strftime("%Y-%m-%d %H:%M:%S",gmtime).' HTML bounce of '.$line.' from '.$from;
exec @logfile;
die $!;


# Extract header (arg 1) from headers string (arg 2)
sub getHeader ($;$)
{
  my ($hdr, $hdrs) = @_;
  $hdrs = $headers unless defined $hdrs;
  $hdrs =~ /^  $hdr:[ \t]*(.*\n(?:  [ \t].*\n)*)/mi;
  my $hline = $1;
  return undef unless defined $hline;
  chomp $hline;
  $hline =~ s/\n[ \t]+/ /go;
  return $hline;
}


# Scan stdin for MIME type regexp $re_* in some multipart/*
# Args as for getHeader
# Returns '' if nothing found, else the interesting MIME type
sub scanbody ($;$@)
{
  local ($mime, $hdrs) = @_;
  local ($line, $boundary, $eof, $ct, $ismime);
  local @search = @_[2..$#_];
  my $i;
  $hdrs = $headers unless defined $hdrs;

  $mime =~ $re_multipart;
  $i = $1;
  push @search, $i if defined $re_secondary{$i} && not grep { $_ eq $i } @search;

  # Get the boundary string
  return '' unless $mime =~ /\bboundary=(".*"|'.*'|[^ \t]*)/io;
  $boundary = $1;
  $boundary =~ s/^.(.*).$/$1/o if substr ($boundary, 0, 1) =~ /'|"/o;
  $boundary = '--'.$boundary;
  $eof = $boundary.'--';

  while (1) {
    # Look for a boundary string
    while ($line = <>) {
      $line =~ s/[\r\n]+$//o;
      return '' if $line eq $eof;
      last if $line eq $boundary;
    }
    return '' unless defined $line;

    # Now slurp in the chunk headers
    undef $ct;
    undef $ismime;
    undef $mime;
    $hdrs = '';
    while ($line = <>) {
      last if $line =~ /^\r?\n?$/o;
      $hdrs .= '  '.$line;
    }
    return '' unless defined $line;

    # Content-Type header?
    $mime = getHeader ('Content-Type', $hdrs);
    return unless defined $mime;

    # Return now if we've found something interesting
    foreach $i (@search) {
      return $1 if $mime =~ $re_secondary{$i};
    }

    # Recursion if we've found multipart
    # Drop straight out if we've found something interesting
    if ($mime =~ $re_multipart) {
      $line = scanbody ($mime, $hdrs, @search);
      return $line if $line ne '';
    }
  }
}
